/*
** pork_aim.c - Interface to libfaim.
** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
**
** This file is heavily based on src/protocols/oscar/oscar.c from gaim,
** which is more or less to say that chunks of code have been mostly
** copied from there.  In turn, I think that oscar.c file was mostly lifed
** from faimtest.  Any copyrights there apply here.
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>

#include <libfaim/aim.h>
#include <libfaim/faimconfig.h>

#include <pork_missing.h>
#include <pork_util.h>
#include <pork_list.h>
#include <pork_buddy.h>
#include <pork_set.h>
#include <pork_swindow.h>
#include <pork_imwindow.h>
#include <pork_acct.h>
#include <pork_cstr.h>
#include <pork_misc.h>
#include <pork_html.h>
#include <pork_events.h>
#include <pork_screen_cmd.h>
#include <pork_aim.h>
#include <pork_io.h>
#include <pork_format.h>

static u_int32_t pork_caps = AIM_CAPS_CHAT;

static char *msgerrreason[] = {
	"Invalid error",
	"Invalid SNAC",
	"Rate to host",
	"Rate to client",
	"Not logged in",
	"Service unavailable",
	"Service not defined",
	"Obsolete SNAC",
	"Not supported by host",
	"Not supported by client",
	"Refused by client",
	"Reply too big",
	"Responses lost",
	"Request denied",
	"Busted SNAC payload",
	"Insufficient rights",
	"In local permit/deny",
	"Sender's warning level is too high",
	"Receiver's warning level is too high",
	"User temporarily unavailable",
	"No match",
	"List overflow",
	"Request ambiguous",
	"Queue full",
	"Not while on AOL"
};

static void aim_debug(	aim_session_t *session __notused,
						int level __notused,
						const char *format __notused,
						va_list va __notused)
{
#ifdef ENABLE_DEBUGGING
	char buf[8192];

	vsnprintf(buf, sizeof(buf), format, va);
	screen_err_msg("LIBFAIM DEBUG: lev: %d: %s", level, buf);
#endif
}

static void autoresp_destroy_cb(void *param __notused, void *data) {
	struct autoresp *autoresp = (struct autoresp *) data;

	free(autoresp->name);
	free(autoresp);
}

static int autoresp_compare_cb(void *l, void *r) {
	char *name = (char *) l;
	struct autoresp *autoresp = (struct autoresp *) r;

	return (strcasecmp(name, autoresp->name));
}

static void aim_send_autoresp(struct aim_priv *priv, char *name) {
	dlist_t *node;
	u_int32_t hash_val = string_hash(name, priv->autoreply.order);

	node = hash_find(&priv->autoreply, name, hash_val);
	if (node == NULL) {
		struct autoresp *autoresp;

		autoresp = xcalloc(1, sizeof(*autoresp));
		autoresp->name = xstrdup(name);
		autoresp->last = time(NULL);

		hash_add(&priv->autoreply, autoresp, hash_val);
		aim_send_msg_auto(priv, name, priv->away_msg);
	} else {
		time_t time_now = time(NULL);
		struct autoresp *autoresp = node->data;

		/*
		** Only send someone an auto-reply every 10 minutes.
		** XXX: maybe this should be a configurable value.
		*/

		if (autoresp->last + 600 <= time_now) {
			autoresp->last = time_now;
			aim_send_msg_auto(priv, name, priv->away_msg);
		}
	}
}

static void chat_free_buddy(void *param __notused, void *data) {
	struct buddy *buddy = (struct buddy *) data;

	free(buddy->nname);
	free(buddy->name);
	free(buddy);
}

static int chat_user_find_cb(void *l, void *r) {
	char *user = (char *) l;
	struct buddy *buddy = (struct buddy *) r;
	char nname[NUSER_LEN];

	normalize(user, nname, sizeof(nname));
	return (strcasecmp(nname, buddy->nname));
}

static void aim_free_chat(struct aim_priv *priv, dlist_t *node) {
	struct chatroom *chat = node->data;

	priv->chat_list = dlist_remove(priv->chat_list, node);

	if (chat->win) {
		char buf[256];

		make_chat_info(opt_get_str(OPT_FORMAT_CHAT_INFO),
			buf, sizeof(buf), priv->username,
			opt_get_str(OPT_FORMAT_CHAT_LEAVE),
			chat->title, chat->exchange);

		screen_win_msg(chat->win, 1, 1, 1, "%s", buf);
	}

	pork_io_del(chat->conn);
	aim_conn_kill(&priv->aim_session, &chat->conn);

	dlist_destroy(chat->user_list, NULL, chat_free_buddy);
	free(chat->name);
	free(chat->title);
	free(chat);
}

static int aim_chat_user_ignored(char *user, struct chatroom *chat) {
	dlist_t *node;
	struct buddy *buddy;

	node = dlist_find(chat->user_list, user, chat_user_find_cb);
	if (node == NULL)
		return (-1);

	buddy = node->data;
	return (buddy->ignore);
}

static char *get_chatname(const char *orig) {
	char *p;
	char *s;
	char *ret;

	if (orig == NULL)
		return (xstrdup(orig));

	p = strrchr(orig, '-');
	if (p == NULL)
		return (xstrdup(orig));

	s = xstrdup(++p);
	ret = s;

	/*
	** The length of s will always be less than or equal to the length of p.
	*/

	while (*p != '\0') {
		if (*p != '%')
			*s = *p++;
		else {
			char buf[3];

			xstrncpy(buf, ++p, sizeof(buf));
			p += 2;

			*s = (char) strtol(buf, NULL, 16);
		}

		s++;
	}

	*s = '\0';
	return (ret);
}

static void aim_chat_add_user(aim_userinfo_t *userinfo, struct chatroom *chat) {
	struct buddy *buddy;
	char nname[NUSER_LEN];

	chat->num_users++;

	normalize(userinfo->sn, nname, sizeof(nname));

	buddy = xcalloc(1, sizeof(*buddy));
	buddy->name = xstrdup(userinfo->sn);
	buddy->nname = xstrdup(nname);

	aim_update_buddy(buddy, userinfo);
	chat->user_list = dlist_add_head(chat->user_list, buddy);
}

static int aim_chat_del_user(aim_userinfo_t *info, struct chatroom *chat) {
	dlist_t *node;
	struct buddy *buddy;

	node = dlist_find(chat->user_list, info->sn, chat_user_find_cb);
	if (node == NULL)
		return (-1);

	chat->num_users--;

	buddy = node->data;
	chat->user_list = dlist_remove(chat->user_list, node);

	free(buddy->name);
	free(buddy->nname);
	free(buddy);

	return (0);
}

static void aim_disconnected(struct aim_priv *priv) {
	int was_connected = priv->connected;
	dlist_t *node;

	pork_io_del(priv->bos_conn);
	aim_conn_kill(&priv->aim_session, &priv->bos_conn);

	priv->connected = 0;
	priv->flags = 0;
	priv->warn_level = 0;
	priv->idle_time = 0;
	priv->bos_conn = NULL;

	buddy_offline_all(priv);

	if (do_event(priv->events, EVENT_SIGNOFF, priv->username))
		return;

	if (was_connected) {
		screen_win_msg(cur_window(), 1, 1, 1,
			"The screen name %s has been disconnected", priv->username);

		if (opt_get_bool(OPT_AUTO_RECONNECT)) {
			screen_err_msg("Automatically reconnecting account %s",
				priv->username);

			aim_login(priv);
			return;
		}
	}

	node = pork_acct_find(priv->username);
	if (node != NULL)
		pork_acct_del(node);
}

static void aim_disconnected_chat(	struct aim_priv *priv,
									struct chatroom *chat)
{
	screen_win_msg(chat->win, 1, 1, 1,
		"%s has been disconnected from %%C%s%%x",
		priv->username, chat->name);

	chat->win->chat = NULL;
	chat->win = NULL;

	aim_leave_chatroom(priv, chat->title, chat->exchange);
}

static void aim_conn_event(int fd __notused, u_int32_t cond, void *data) {
	aim_conn_t *conn = (aim_conn_t *) data;
	aim_session_t *session = aim_conn_getsess(conn);
	struct aim_priv *priv = session->aux_data;

	if (cond == IO_COND_DEAD) {
		switch (conn->type) {
			/*
			** You sunk my battleship!
			*/
			case AIM_CONN_TYPE_BOS:
				aim_disconnected(priv);
				break;

			case AIM_CONN_TYPE_CHAT:
				aim_disconnected_chat(priv, conn->priv);
				break;

			case AIM_CONN_TYPE_CHATNAV:
				pork_io_del(conn);
				aim_conn_kill(session, &conn);
				priv->chatnav_conn = NULL;
				break;
		}

		return;
	}

	if (cond & IO_COND_READ) {
		if (aim_get_command(session, conn) >= 0) {
			aim_rxdispatch(session);

			/*
			** If the whole session is terminated (e.g. after a failed
			** login, we can't destroy the account inside the callback
			** (before returning from aim_rxdispatch), or it will crash
			** the client.  Instead, this will have to do.
			*/

			if (session->connlist == NULL) {
				dlist_t *node;

				node = pork_acct_find(priv->username);
				if (node != NULL)
					pork_acct_del(node);

				return;
			}
		}
	}

	if (cond & IO_COND_WRITE)
		aim_tx_flushqueue(session);
}

static int aim_send_buddy_list(	aim_session_t *session,
								aim_conn_t *conn,
								struct aim_priv *priv)
{
	dlist_t *cur;

	cur = priv->buddy_pref->group_list;
	while (cur != NULL) {
		struct bgroup *group = cur->data;
		dlist_t *cur_buddy = group->member_list;

		while (cur_buddy != NULL) {
			struct buddy *buddy = cur_buddy->data;

			if (aim_ssi_itemlist_exists(session->ssi.local, buddy->nname)) {
				char *alias;

				alias = aim_ssi_getalias(session->ssi.local,
							group->name, buddy->nname);

				if (alias != NULL) {
					free(buddy->name);
					buddy->name = alias;
				} else {
					aim_ssi_aliasbuddy(session, conn,
						group->name, buddy->nname, buddy->name);
				}
			} else {
				aim_ssi_addbuddy(session, conn, buddy->nname,
					group->name, buddy->name, NULL, NULL, 0);
			}

			cur_buddy = cur_buddy->next;
		}

		cur = cur->next;
	}

	return (1);
}

static int aim_send_permit_list(aim_session_t *session,
								aim_conn_t *conn,
								struct aim_priv *priv)
{
	if (priv->buddy_pref->permit_list != NULL) {
		dlist_t *cur;

		cur = priv->buddy_pref->permit_list;
		while (cur != NULL) {
			aim_ssi_addpermit(session, conn, (char *) cur->data);
			cur = cur->next;
		}
	}

	return (1);
}

static int aim_send_block_list(	aim_session_t *session,
								aim_conn_t *conn,
								struct aim_priv *priv)
{
	if (priv->buddy_pref->block_list != NULL) {
		dlist_t *cur;

		cur = priv->buddy_pref->block_list;
		while (cur != NULL) {
			aim_ssi_adddeny(session, conn, (char *) cur->data);
			cur = cur->next;
		}
	}

	return (1);
}

/*
** libfaim callback handlers.
*/

static FAIM_CB(aim_conn_dead) {
	aim_conn_t *conn;
	va_list ap;

	va_start(ap, fr);
	conn = va_arg(ap, aim_conn_t *);
	va_end(ap);

	pork_io_dead(conn);
	return (1);
}

static FAIM_CB(aim_recv_bos_rights) {
	va_list ap;
	struct aim_priv *priv = session->aux_data;

	priv->bos_conn = fr->conn;

	va_start(ap, fr);
	priv->rights.max_permit = va_arg(ap, unsigned int);
	priv->rights.max_deny = va_arg(ap, unsigned int);
	va_end(ap);

	aim_ssi_reqrights(session, fr->conn);
	aim_ssi_reqdata(session, fr->conn, session->ssi.timestamp,
		session->ssi.numitems);

	return (1);
}

static FAIM_CB(aim_recv_buddy_rights) {
	va_list ap;
	struct aim_priv *priv = session->aux_data;

	va_start(ap, fr);
	priv->rights.max_buddies = va_arg(ap, unsigned int);
	priv->rights.max_watchers = va_arg(ap, unsigned int);
	va_end(ap);

	return (1);
}

static FAIM_CB(aim_recv_conn_complete) {
	pork_io_set_cond(fr->conn, IO_COND_READ);

	aim_reqpersonalinfo(session, fr->conn);
	aim_bos_reqlocaterights(session, fr->conn);
	aim_bos_reqbuddyrights(session, fr->conn);
	aim_reqicbmparams(session);
	aim_bos_reqrights(session, fr->conn);

	return (1);
}

static FAIM_CB(aim_recv_typing) {
	va_list ap;
	u_int16_t type1;
	u_int16_t type2;
	char *source;
	struct imwindow *conv_window;

	va_start(ap, fr);
	type1 = (u_int16_t) va_arg(ap, unsigned int);
	source = va_arg(ap, char *);
	type2 = (u_int16_t) va_arg(ap, unsigned int);
	va_end(ap);

	conv_window = imwindow_find(source);
	if (conv_window == NULL)
		return (0);

	if (type2 > 2)
		return (0);

	conv_window->typing = type2;
	return (0);
}

static FAIM_CB(aim_recv_err_loc) {
	va_list ap;
	char *dest;
	char *err_str;
	u_int16_t code;
	struct aim_priv *priv = session->aux_data;

	va_start(ap, fr);
	code = (u_int16_t) va_arg(ap, unsigned int);
	dest = va_arg(ap, char *);
	va_end(ap);

	if (code < array_elem(msgerrreason))
		err_str = msgerrreason[code];
	else
		err_str = "Reason unknown";

	if (priv->whois_hack)
		priv->whois_hack = 0;
	else {
		screen_output("User information for is %s unavailable: %s",
			dest, err_str);
	}

	return (1);
}

static FAIM_CB(aim_recv_err_msg) {
	va_list ap;
	char *dest;
	char *err_str;
	u_int16_t code;
	struct aim_priv *priv = session->aux_data;

	va_start(ap, fr);
	code = (u_int16_t) va_arg(ap, unsigned int);
	dest = va_arg(ap, char *);
	va_end(ap);

	if (code < array_elem(msgerrreason))
		err_str = msgerrreason[code];
	else
		err_str = "Reason unknown";

	screen_win_target_msg(dest, 0,
		"%s's message to %s was not sent: %s", priv->username, dest, err_str);

	return (1);
}

static FAIM_CB(aim_recv_err_other) {
	va_list ap;
	u_int16_t code;
	char *err_str;

	va_start(ap, fr);
	code = (u_int16_t) va_arg(ap, unsigned int);
	va_end(ap);

	if (code < array_elem(msgerrreason))
		err_str = msgerrreason[code];
	else
		err_str = "Reason unknown";

	screen_err_msg("AIM Error: %s", err_str);
	return (1);
}

static FAIM_CB(aim_recv_evil) {
	va_list ap;
	u_int16_t warn_level;
	aim_userinfo_t *userinfo;
	struct aim_priv *priv = session->aux_data;
	char buf[4096];

	va_start(ap, fr);
	warn_level = va_arg(ap, unsigned int);
	userinfo = va_arg(ap, aim_userinfo_t *);
	va_end(ap);

	warn_level = (float) warn_level / 10;
	if (warn_level <= priv->warn_level) {
		priv->warn_level = warn_level;
		return (1);
	}

	priv->warn_level = warn_level;

	if (do_event(priv->events, EVENT_RECV_WARN, userinfo->sn, warn_level,
		priv->username))
	{
		return (1);
	}

	if (blank_str(userinfo->sn)) {
		make_warning(opt_get_str(OPT_FORMAT_WARN),
			buf, sizeof(buf), priv->username,
			opt_get_str(OPT_TEXT_WARN_ANONYMOUS), warn_level);

		screen_output("%s", buf);
	} else {
		struct buddy *buddy;
		char *name;

		buddy = buddy_find(priv, userinfo->sn);
		if (buddy != NULL)
			name = buddy->name;
		else
			name = userinfo->sn;

		make_warning(opt_get_str(OPT_FORMAT_WARN),
			buf, sizeof(buf), priv->username,
			name, warn_level);

		screen_win_target_msg(userinfo->sn, 1, "%s", buf);
	}

	return (1);
}

static FAIM_CB(aim_recv_icbm_param_info) {
	va_list ap;
	struct aim_icbmparameters *params;

	va_start(ap, fr);
	params = va_arg(ap, struct aim_icbmparameters *);
	va_end(ap);

	params->flags = 0x0000000b;
	params->maxmsglen = 8000;
	params->minmsginterval = 0;

	aim_seticbmparam(session, params);
	return (1);
}


static int parse_im_chan1(	aim_session_t *session,
							aim_conn_t *conn,
							aim_userinfo_t *userinfo,
							struct aim_incomingim_ch1_args *args)
{
	char *buf;
	char *final_msg;
	int autoresp = 0;
	struct imwindow *conv_window;
	struct aim_priv *priv = session->aux_data;
	struct buddy *buddy;
	char *name;

	(void) session;
	(void) conn;

	if (args->icbmflags & AIM_IMFLAGS_AWAY)
		autoresp = 1;

	if (args->icbmflags & AIM_IMFLAGS_UNICODE) {
		int i;
		char *msg;

		buf = xmalloc(args->msglen + 1);
		msg = buf;

		for (i = 0 ; i < args->msglen ; i += 2) {
			u_int16_t uc_char;

			uc_char = ((args->msg[i] & 0xff) << 8) | (args->msg[i + 1] & 0xff);

			if (uc_char > 255)
				*msg++ = (char) uc_char;
			else
				*msg++ = '?';
		}

		*msg = '\0';
	} else
		buf = xstrdup(args->msg);

	final_msg = strip_html(buf);
	free(buf);

	buddy = buddy_find(priv, userinfo->sn);
	if (buddy != NULL)
		name = buddy->name;
	else
		name = userinfo->sn;

	conv_window = imwindow_find(userinfo->sn);
	if (conv_window == NULL) {
		conv_window = screen_new_window(priv, userinfo->sn, name);
		if (conv_window == NULL)
			return (-1);
	}

	conv_window->typing = 0;

	if (!do_event(priv->events, EVENT_RECV_IM, userinfo->sn, autoresp,
		final_msg, priv->username))
	{
		imwindow_recv_im(conv_window, name, autoresp, final_msg);
		if (priv->away_msg != NULL && !autoresp) {
			if (opt_get_bool(OPT_AUTOSEND_AWAY))
				aim_send_autoresp(priv, userinfo->sn);
		}
	}

	free(final_msg);
	return (1);
}

static int parse_im_chan2(	aim_session_t *session,
							aim_conn_t *conn,
							aim_userinfo_t *userinfo,
							struct aim_incomingim_ch2_args *args)
{
	struct aim_priv *priv = session->aux_data;

	if (args->reqclass & AIM_CAPS_CHAT) {
		u_int16_t exchange = args->info.chat.roominfo.exchange;
		char *chat_name;

		/*
		** They accepted our invitation.
		*/

		if (args->info.chat.roominfo.name == NULL && args->msg == NULL) {
			struct chatroom *chat = conn->priv;

			if (chat != NULL) {
				char buf[256];

				aim_chat_add_user(userinfo, chat);

				if (do_event(priv->events, EVENT_RECV_CHAT_JOIN, chat->name,
					chat->title, chat->exchange, userinfo->sn, priv->username))
				{
					return (1);
				}

				make_chat_info(opt_get_str(OPT_FORMAT_CHAT_INFO),
					buf, sizeof(buf), userinfo->sn,
					opt_get_str(OPT_FORMAT_CHAT_JOIN),
					chat->title, chat->exchange);

				screen_win_msg(chat->win, 1, 1, 1, "%s", buf);
			}

			return (1);
		}

		chat_name = get_chatname(args->info.chat.roominfo.name);
		if (chat_name == NULL)
			chat_name = xstrdup(args->info.chat.roominfo.name);

		if (!do_event(priv->events, EVENT_RECV_CHAT_INVITE,
			args->info.chat.roominfo.name, chat_name, exchange,
			userinfo->sn, (char *) args->msg, priv->username))
		{
			char buf[256];

			make_chat_arginfo(opt_get_str(OPT_FORMAT_CHAT_INVITE),
				buf, sizeof(buf), userinfo->sn, chat_name,
				exchange, priv->username, args->msg);

			screen_win_msg(cur_window(), 1, 1, 1, "%s", buf);
		}

		free(chat_name);
		return (1);
	}

	return (0);
}

static FAIM_CB(aim_recv_msg) {
	va_list ap;
	int channel;
	aim_userinfo_t *userinfo;
	int ret = 0;

	va_start(ap, fr);
	channel = va_arg(ap, int);
	userinfo = va_arg(ap, aim_userinfo_t *);

	switch (channel) {
		case 1:
		{
			struct aim_incomingim_ch1_args *args;

			args = va_arg(ap, struct aim_incomingim_ch1_args *);
			ret = parse_im_chan1(session, fr->conn, userinfo, args);

			break;
		}

		case 2:
		{
			struct aim_incomingim_ch2_args *args;

			args = va_arg(ap, struct aim_incomingim_ch2_args *);
			ret = parse_im_chan2(session, fr->conn, userinfo, args);

			break;
		}

		/* XXX - fix this for directIM */

	}

	va_end(ap);
	return (ret);
}

static FAIM_CB(aim_recv_locrights) {
	va_list ap;
	u_int16_t max_len;
	struct aim_priv *priv = session->aux_data;

	va_start(ap, fr);
	max_len = va_arg(ap, int);
	va_end(ap);

	priv->rights.max_profile_len = max_len;
	priv->rights.max_away_len = max_len;

	if (priv->profile != NULL) {
		char *profile_html;
		size_t profile_len;

		profile_html = text2html(priv->profile);
		profile_len = strlen(profile_html);

		aim_bos_setprofile(session, fr->conn, "us-ascii", profile_html,
			profile_len, NULL, NULL, 0, pork_caps);
	}

	if (priv->away_msg != NULL) {
		char *away_html;
		size_t away_len;

		away_html = text2html(priv->away_msg);
		away_len = strlen(away_html);

		aim_bos_setprofile(session, fr->conn, NULL, NULL, 0,
			"us-ascii", away_html, away_len, pork_caps);
	}

	if (priv->profile == NULL && priv->away_msg == NULL) {
		aim_bos_setprofile(session, fr->conn,
			NULL, NULL, 0, NULL, NULL, 0, pork_caps);
	}

	return (1);
}

static FAIM_CB(aim_recv_mem_request) {
	screen_err_msg("Unhandled event: aim_recv_mem_request");
	return (1);
}

static FAIM_CB(aim_recv_clientauto) {
	/* XXX - ignore this for now */
	return (1);
}

static FAIM_CB(aim_recv_missed) {
	va_list ap;
	u_int16_t chan;
	u_int16_t num_missed;
	u_int16_t reason;
	aim_userinfo_t *userinfo;
	char buf[1024];
	struct aim_priv *priv = session->aux_data;

	va_start(ap, fr);
	chan = (u_int16_t) va_arg(ap, unsigned int);
	userinfo = va_arg(ap, aim_userinfo_t *);
	num_missed = (u_int16_t) va_arg(ap, unsigned int);
	reason = (u_int16_t) va_arg(ap, unsigned int);
	va_end(ap);

	switch(reason) {
		case 0:
			snprintf(buf, sizeof(buf),
				num_missed == 1 ?
					"%s missed %d message from %s because it was invalid" :
					"%s missed %d messages from %s because they were invalid",
				priv->username, num_missed, userinfo->sn);
			break;

		case 1:
			snprintf(buf, sizeof(buf),
				num_missed == 1 ?
					"%s missed %d message from %s because it was too large" :
					"%s missed %d messages from %s because they were too large",
				priv->username, num_missed, userinfo->sn);
			break;

		case 2:
			snprintf(buf, sizeof(buf),
				num_missed == 1 ?
					"%s missed %d message from %s because the rate limit has been exceeded" :
					"%s missed %d messages from %s because the rate limit has been exceeded",
				priv->username, num_missed, userinfo->sn);
			break;

		case 3:
			snprintf(buf, sizeof(buf),
				num_missed == 1 ?
					"%s missed %d message from %s because the sender's warning level is too high" :
					"%s missed %d messages from %s because the sender's warning level is too high",
				priv->username, num_missed, userinfo->sn);
			break;

		case 4:
			snprintf(buf, sizeof(buf),
				num_missed == 1 ?
					"%s missed %d message from %s because your warning level is too high" :
					"%s missed %d messages from %s because your warning level is too high",
				priv->username, num_missed, userinfo->sn);
			break;

		default:
			snprintf(buf, sizeof(buf),
				num_missed == 1 ?
					"%s missed %d message from %s for unknown reasons" :
					"%s missed %d messages from %s for unknown reasons",
				priv->username, num_missed, userinfo->sn);
			break;
	}

	return (1);
}

static FAIM_CB(aim_recv_selfinfo) {
	va_list ap;
	aim_userinfo_t *info;
	struct aim_priv *priv = session->aux_data;

	va_start(ap, fr);
	info = va_arg(ap, aim_userinfo_t *);
	va_end(ap);

	priv->warn_level = (float) info->warnlevel / 10;
	priv->idle_time = info->idletime;

	return (1);
}

static FAIM_CB(aim_recv_offgoing) {
	va_list ap;
	aim_userinfo_t *userinfo;
	struct buddy *buddy;
	struct aim_priv *priv = session->aux_data;
	struct imwindow *conv_window;

	va_start(ap, fr);
	userinfo = va_arg(ap, aim_userinfo_t *);
	va_end(ap);

	if (userinfo == NULL)
		return (-1);

	buddy = buddy_find(priv, userinfo->sn);
	if (buddy == NULL) {
		screen_err_msg("offgoing for unknown buddy %s", userinfo->sn);
		return (1);
	}

	/*
	** I'm not exactly sure why this happens.
	*/
	if (buddy->status == STATUS_OFFLINE)
		return (1);

	if (do_event(priv->events, EVENT_BUDDY_SIGNOFF, userinfo->sn,
		priv->username))
	{
		return (1);
	}

	/*
	** If there's a conversation window with this user,
	** print the signoff message inside that window.
	**
	** If there's no conversation window, but notify
	** is set, print the signoff message in the current window.
	*/

	conv_window = imwindow_find(userinfo->sn);
	if (conv_window != NULL) {
		screen_win_target_msg(userinfo->sn, 1, "%s has signed off",
			buddy->name);
	} else {
		if (buddy->notify) {
			screen_win_target_msg(userinfo->sn, 1, "%s has signed off",
				buddy->name);
		}
	}

	buddy_offline(priv, buddy);
	return (1);
}

static FAIM_CB(aim_recv_oncoming) {
	va_list ap;
	aim_userinfo_t *userinfo;
	struct buddy *buddy;
	struct aim_priv *priv = session->aux_data;
	char *status_msg = NULL;

	va_start(ap, fr);
	userinfo = va_arg(ap, aim_userinfo_t *);
	va_end(ap);

	if (userinfo == NULL)
		return (-1);

	buddy = buddy_find(priv, userinfo->sn);
	if (buddy == NULL) {
		screen_err_msg("oncoming for unknown buddy %s", userinfo->sn);
		return (1);
	}

	if (buddy->status == STATUS_OFFLINE) {
		if (do_event(priv->events, EVENT_BUDDY_SIGNON, userinfo->sn,
			priv->username))
		{
			return (1);
		}

		buddy_online(priv, buddy, userinfo);

		if (opt_get_bool(OPT_SHOW_BUDDY_SIGNOFF))
			status_msg = "has signed on";

	} else {
		int was_away = buddy->status == STATUS_AWAY;
		int is_away = (userinfo->flags & AIM_FLAG_AWAY) != 0;

		if (was_away != is_away) {
			if (was_away) {
				if (do_event(priv->events, EVENT_BUDDY_BACK, userinfo->sn,
					priv->username))
				{
					return (1);
				}

				if (opt_get_bool(OPT_SHOW_BUDDY_AWAY))
					status_msg = "is no longer away";
			} else {
				if (do_event(priv->events, EVENT_BUDDY_AWAY, userinfo->sn,
					priv->username))
				{
					return (1);
				}

				if (opt_get_bool(OPT_SHOW_BUDDY_AWAY))
					status_msg = "is now away";
			}
		} else if (buddy->idle_time && !userinfo->idletime) {
			if (do_event(priv->events, EVENT_BUDDY_UNIDLE, userinfo->sn,
				priv->username))
			{
				return (1);
			}

			if (opt_get_bool(OPT_SHOW_BUDDY_IDLE))
				status_msg = "is no longer idle";

		} else if (!buddy->idle_time && userinfo->idletime) {
			if (do_event(priv->events, EVENT_BUDDY_IDLE, userinfo->sn,
				userinfo->idletime, priv->username))
			{
				return (1);
			}

			if (opt_get_bool(OPT_SHOW_BUDDY_IDLE))
				status_msg = "is now idle";

		}

		buddy_update(priv, buddy, userinfo);
	}

	/*
	** If there's a conversation window with this user,
	** print the signoff message inside that window.
	**
	** If there's no conversation window, but notify
	** is set, print the signoff message in the current window.
	*/

	if (status_msg != NULL) {
		struct imwindow *conv_window = imwindow_find(userinfo->sn);
		if (conv_window != NULL) {
			screen_win_target_msg(userinfo->sn, 1, "%s %s",
				buddy->name, status_msg);
		} else {
			if (buddy->notify) {
				screen_win_target_msg(userinfo->sn, 1, "%s %s",
					buddy->name, status_msg);
			}
		}
	}

	return (1);
}

static FAIM_CB(aim_recv_rate_change) {
	struct aim_priv *priv = session->aux_data;
	va_list ap;
	u_int16_t rate_code;
	u_int16_t rate_class;
	u_int32_t window_size;
	u_int32_t rate_clear;
	u_int32_t alert;
	u_int32_t limit;
	u_int32_t disconnect;
	u_int32_t current_avg;
	u_int32_t max_avg;

	va_start(ap, fr);

	rate_code = (u_int16_t) va_arg(ap, unsigned int);
	rate_class = (u_int16_t) va_arg(ap, unsigned int);
	window_size = (u_int32_t) va_arg(ap, u_int32_t);
	rate_clear = (u_int32_t) va_arg(ap, u_int32_t);
	alert = (u_int32_t) va_arg(ap, u_int32_t);
	limit = (u_int32_t) va_arg(ap, u_int32_t);
	disconnect = (u_int32_t) va_arg(ap, u_int32_t);
	current_avg = (u_int32_t) va_arg(ap, u_int32_t);
	max_avg = (u_int32_t) va_arg(ap, u_int32_t);

	va_end(ap);

	switch (rate_code) {
		case AIM_RATE_CODE_CHANGE:
			if (current_avg >= rate_clear)
				aim_conn_setlatency(fr->conn, 0);
			break;

		case AIM_RATE_CODE_WARNING:
			aim_conn_setlatency(fr->conn, window_size / 4);
			break;

		case AIM_RATE_CODE_LIMIT:
			aim_conn_setlatency(fr->conn, window_size / 2);
			screen_err_msg("The last message from %s was not sent because you are over the rate limit. Please wait 10 seconds, and then try again", priv->username);
			break;

		case AIM_RATE_CODE_CLEARLIMIT:
			aim_conn_setlatency(fr->conn, 0);
			break;
	}

	return (1);
}

static FAIM_CB(aim_recv_chatnav_info) {
	va_list ap;
	u_int16_t type;
	struct aim_priv *priv = session->aux_data;

	va_start(ap, fr);
	type = (u_int16_t) va_arg(ap, unsigned int);

	switch (type) {
		case 0x0002:
		{
			u_int8_t max_rooms;
			int exchange_count;
			struct aim_chat_exchangeinfo *exchanges;
			dlist_t *cur;

			max_rooms = (u_int8_t) va_arg(ap, unsigned int);
			exchange_count = va_arg(ap, int);
			exchanges = va_arg(ap, struct aim_chat_exchangeinfo *);

			cur = priv->chat_create_list;
			while (cur != NULL) {
				struct chatroom_info *chat = cur->data;
				dlist_t *next = cur->next;

				aim_chatnav_createroom(session, fr->conn, chat->name,
					chat->exchange);

				priv->chat_create_list =
					dlist_remove(priv->chat_create_list, cur);

				free(chat->name);
				free(chat);

				cur = next;
			}

			break;
		}

		case 0x0008:
		{
			char *fqcn;
			char *name;
			char *ck;
			u_int16_t instance;
			u_int16_t flags;
			u_int16_t maxmsglen;
			u_int16_t maxoccupancy;
			u_int16_t unknown;
			u_int16_t exchange;
			u_int8_t createperms;
			u_int32_t createtime;

			fqcn = va_arg(ap, char *);
			instance = (u_int16_t) va_arg(ap, unsigned int);
			exchange = (u_int16_t) va_arg(ap, unsigned int);
			flags = (u_int16_t) va_arg(ap, unsigned int);
			createtime = va_arg(ap, u_int32_t);
			maxmsglen = (u_int16_t) va_arg(ap, unsigned int);
			maxoccupancy = (u_int16_t) va_arg(ap, unsigned int);
			createperms = (u_int8_t) va_arg(ap, unsigned int);
			unknown = (u_int16_t) va_arg(ap, unsigned int);
			name = va_arg(ap, char *);
			ck = va_arg(ap, char *);

			aim_chat_join(session, priv->bos_conn, exchange, ck, instance);
			break;
		}
	}

	va_end(ap);
	return (1);
}

static FAIM_CB(aim_recv_chat_join) {
	struct aim_priv *priv = session->aux_data;
	struct chatroom *chat = fr->conn->priv;
	int num_users;
	aim_userinfo_t *info;
	va_list ap;
	u_int32_t old_num_users = chat->num_users;
	int i;
	int ret = 0;

	va_start(ap, fr);
	num_users = va_arg(ap, int);
	info = va_arg(ap, aim_userinfo_t *);
	va_end(ap);

	for (i = 0 ; i < num_users ; i++) {
		int event_ret;

		aim_chat_add_user(&info[i], chat);

		event_ret = do_event(priv->events, EVENT_RECV_CHAT_JOIN, chat->name,
			chat->title, chat->exchange, info[i].sn, priv->username);

		if (event_ret != 0)
			ret = 1;
	}

	if (ret != 0)
		return (1);

	if (old_num_users == 0 && num_users == 1) {
		char buf[256];

		make_chat_arginfo(opt_get_str(OPT_FORMAT_CHAT_CREATE),
			buf, sizeof(buf), priv->username,
			chat->title, chat->exchange, priv->username, chat->name);

		screen_win_msg(chat->win, 1, 1, 1, "%s", buf);
	} else {
		char buf[256];

		if (old_num_users != 0) {
			make_chat_info(opt_get_str(OPT_FORMAT_CHAT_INFO),
				buf, sizeof(buf), info->sn,
				opt_get_str(OPT_FORMAT_CHAT_JOIN),
				chat->title, chat->exchange);

			screen_win_msg(chat->win, 1, 1, 1, "%s", buf);
		} else {
			make_chat_info(opt_get_str(OPT_FORMAT_CHAT_INFO),
				buf, sizeof(buf), priv->username,
				opt_get_str(OPT_FORMAT_CHAT_JOIN),
				chat->title, chat->exchange);

			screen_win_msg(chat->win, 1, 1, 1, "%s", buf);
			aim_chat_print_users(chat->win, chat);
		}
	}

	return (1);
}

static FAIM_CB(aim_recv_chat_leave) {
	char buf[256];
	struct aim_priv *priv = session->aux_data;
	struct chatroom *chat = fr->conn->priv;
	va_list ap;
	int num_users;
	aim_userinfo_t *info;
	char nname[NUSER_LEN];

	va_start(ap, fr);
	num_users = va_arg(ap, int);
	info = va_arg(ap, aim_userinfo_t *);
	va_end(ap);

	/*
	** Sometimes we'll receive a notice that we're leaving
	** a chat room.  I don't know what makes this happen, but
	** it means that we've been disconnected somehow.
	*/

	normalize(info->sn, nname, sizeof(nname));
	if (!strcmp(nname, priv->username)) {
		aim_disconnected_chat(priv, chat);
		return (1);
	}

	aim_chat_del_user(info, chat);

	if (do_event(priv->events, EVENT_RECV_CHAT_LEAVE, chat->name, chat->title,
		chat->exchange, info->sn, priv->username))
	{
		return (1);
	}

	make_chat_info(opt_get_str(OPT_FORMAT_CHAT_INFO),
		buf, sizeof(buf), info->sn,
		opt_get_str(OPT_FORMAT_CHAT_LEAVE),
		chat->title, chat->exchange);
	screen_win_msg(chat->win, 1, 1, 1, "%s", buf);

	return (1);
}

static FAIM_CB(aim_recv_chat_info_update) {
	va_list ap;
	aim_userinfo_t *userinfo;
	struct aim_chat_roominfo *roominfo;
	char *roomname;
	int usercount;
	char *roomdesc;
	u_int16_t unknown_c9, unknown_d2, unknown_d5, maxmsglen, maxvisiblemsglen;
	u_int32_t creationtime;
	struct chatroom *chat = fr->conn->priv;

	va_start(ap, fr);
	roominfo = va_arg(ap, struct aim_chat_roominfo *);
	roomname = va_arg(ap, char *);
	usercount = va_arg(ap, int);
	userinfo = va_arg(ap, aim_userinfo_t *);
	roomdesc = va_arg(ap, char *);
	unknown_c9 = (u_int16_t) va_arg(ap, unsigned int);
	creationtime = (u_int32_t) va_arg(ap, u_int32_t);
	maxmsglen = (u_int16_t) va_arg(ap, unsigned int);
	unknown_d2 = (u_int16_t) va_arg(ap, unsigned int);
	unknown_d5 = (u_int16_t) va_arg(ap, unsigned int);
	maxvisiblemsglen = (u_int16_t) va_arg(ap, unsigned int);
	va_end(ap);

	chat->max_msg_len = maxmsglen;
	chat->max_visible_len = maxvisiblemsglen;
	chat->created = (time_t) creationtime;

	return (1);
}

static FAIM_CB(aim_recv_chat_msg) {
	va_list ap;
	aim_userinfo_t *info;
	char *msg;
	struct aim_priv *priv = session->aux_data;
	struct chatroom *chat = fr->conn->priv;

	va_start(ap, fr);
	info = va_arg(ap, aim_userinfo_t *);
	msg = va_arg(ap, char *);
	va_end(ap);

	msg = strip_html(msg);
	if (do_event(priv->events, EVENT_RECV_CHAT_MSG, chat->name,
		chat->title, chat->exchange, info->sn, msg, priv->username))
	{
		free(msg);
		return (0);
	}

	if (!aim_chat_user_ignored(info->sn, chat))
		imwindow_recv_im(chat->win, info->sn, 0, msg);

	free(msg);
	return (1);
}

static FAIM_CB(recv_chat_conn) {
	pork_io_set_cond(fr->conn, IO_COND_READ);

	aim_conn_addhandler(session, fr->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_ERROR, aim_recv_err_other, 0);
	aim_conn_addhandler(session, fr->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_USERJOIN, aim_recv_chat_join, 0);
	aim_conn_addhandler(session, fr->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_USERLEAVE, aim_recv_chat_leave, 0);
	aim_conn_addhandler(session, fr->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_ROOMINFOUPDATE, aim_recv_chat_info_update, 0);
	aim_conn_addhandler(session, fr->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_INCOMINGMSG, aim_recv_chat_msg, 0);

	aim_clientready(session, fr->conn);
	return (1);
}

static FAIM_CB(recv_chatnav_conn) {
	struct aim_priv *priv = session->aux_data;

	pork_io_set_cond(fr->conn, IO_COND_READ);

	aim_conn_addhandler(session, fr->conn, AIM_CB_FAM_CTN, AIM_CB_CTN_ERROR, aim_recv_err_other, 0);
	aim_conn_addhandler(session, fr->conn, AIM_CB_FAM_CTN, AIM_CB_CTN_INFO, aim_recv_chatnav_info, 0);

	aim_clientready(session, fr->conn);
	aim_chatnav_reqrights(session, fr->conn);

	priv->chatnav_conn = fr->conn;

	return (1);
}

static FAIM_CB(aim_recv_redirect) {
	struct aim_redirect_data *redirect;
	va_list ap;
	struct aim_priv *priv = session->aux_data;

	va_start(ap, fr);
	redirect = va_arg(ap, struct aim_redirect_data *);
	va_end(ap);

	switch (redirect->group) {
		case AIM_CONN_TYPE_AUTH:
			break;

		case AIM_CONN_TYPE_EMAIL:
			break;

		case AIM_CONN_TYPE_CHATNAV:
		{
			aim_conn_t *chatnav;

			chatnav = aim_newconn(session, AIM_CONN_TYPE_CHATNAV, redirect->ip);
			if (chatnav == NULL || chatnav->fd == -1) {
				if (chatnav != NULL)
					aim_conn_kill(session, &chatnav);
				screen_err_msg("Unable to connect to the chatnav server");
				return (0);
			}

			pork_io_add(chatnav->fd, IO_COND_RW, chatnav, chatnav,
				aim_conn_event);

			aim_conn_addhandler(session, chatnav, AIM_CB_FAM_SPECIAL, AIM_CB_SPECIAL_CONNDEAD, aim_conn_dead, 0);
			aim_conn_addhandler(session, chatnav, AIM_CB_FAM_SPECIAL, AIM_CB_SPECIAL_CONNINITDONE, recv_chatnav_conn, 0);
			aim_sendcookie(session, chatnav, redirect->cookielen, redirect->cookie);
			break;
		}

		case AIM_CONN_TYPE_CHAT:
		{
			aim_conn_t *chat_conn;
			struct chatroom *chat;
			struct imwindow *win;
			char buf[128];

			chat_conn = aim_newconn(session, AIM_CONN_TYPE_CHAT, redirect->ip);
			if (chat_conn == NULL || chat_conn->fd == -1) {
				screen_err_msg("Unable to connect to the chat server");
				return (0);
			}

			pork_io_add(chat_conn->fd, IO_COND_RW, chat_conn, chat_conn,
				aim_conn_event);

			chat = xcalloc(1, sizeof(*chat));
			chat->name = xstrdup(redirect->chat.room);
			chat->title = get_chatname(chat->name);
			chat->exchange = redirect->chat.exchange;
			chat->instance = redirect->chat.instance;
			chat->conn = chat_conn;
			chat_conn->priv = chat;

			priv->chat_list = dlist_add_head(priv->chat_list, chat);

			snprintf(buf, sizeof(buf), "%s/%d", chat->title, chat->exchange);
			win = imwindow_find_chat_target(priv, buf);
			if (win != NULL) {
				win->chat = chat;
				chat->win = win;
			}

			aim_conn_addhandler(session, chat_conn, AIM_CB_FAM_SPECIAL, AIM_CB_SPECIAL_CONNDEAD, aim_conn_dead, 0);
			aim_conn_addhandler(session, chat_conn, AIM_CB_FAM_SPECIAL, AIM_CB_SPECIAL_CONNINITDONE, recv_chat_conn, 0);
			aim_sendcookie(session, chat_conn, redirect->cookielen, redirect->cookie);
			break;
		}
	}

	return (1);
}

static FAIM_CB(aim_recv_search_error) {
	va_list ap;
	char *address;
	struct aim_priv *priv = session->aux_data;

	va_start(ap, fr);
	address = va_arg(ap, char *);
	va_end(ap);

	if (do_event(priv->events, EVENT_RECV_SEARCH_RESULT, address, NULL,
		priv->username))
	{
		return (1);
	}

	screen_output("No results were found for %s", address);
	return (1);
}

static FAIM_CB(aim_recv_search_reply) {
	va_list ap;
	char *address;
	char *usernames;
	char buf[2048];
	int i = 0;
	size_t len = sizeof(buf);
	int j;
	int num;
	int ret;
	struct aim_priv *priv = session->aux_data;

	va_start(ap, fr);
	address = va_arg(ap, char *);
	num = va_arg(ap, int);
	usernames = va_arg(ap, char *);
	va_end(ap);

	if (num < 1) {
		if (do_event(priv->events, EVENT_RECV_SEARCH_RESULT, address, NULL,
			priv->username))
		{
			return (1);
		}

		screen_output("No results were found for %s", address);
		return (1);
	}

	ret = snprintf(buf, len, "%s has registered the following screen names: %s",
			address, &usernames[0]);
	len -= ret;
	i += ret;

	for (j = 1 ; j < num ; j++) {
		ret = snprintf(&buf[i], len, ", %s", &usernames[j * (MAXSNLEN + 1)]);
		len -= ret;
		i += ret;
	}

	if (do_event(priv->events, EVENT_RECV_SEARCH_RESULT, address, buf,
		priv->username))
	{
		return (1);
	}

	screen_output("%s", buf);
	return (1);
}

static FAIM_CB(aim_ssi_recv_rights) {
	struct aim_priv *priv = session->aux_data;
	int num;
	u_int16_t *items;
	va_list ap;

	va_start(ap, fr);
	num = va_arg(ap, int);
	items = va_arg(ap, u_int16_t *);
	va_end(ap);

	if (num >= 0)
		priv->rights.max_buddies = items[0];

	if (num >= 1)
		priv->rights.max_groups = items[1];

	if (num >= 2)
		priv->rights.max_permit = items[2];

	if (num >= 3)
		priv->rights.max_deny = items[3];

	return (1);
}

static FAIM_CB(aim_ssi_recv_list) {
	struct aim_ssi_item *cur;
	struct aim_priv *priv = session->aux_data;

	aim_ssi_enable(session, fr->conn);
	aim_ssi_cleanlist(session, fr->conn);

	/*
	** Merge the buddy list on the server with the locally
	** stored buddy list.  By this point, the local
	** buddy list will have already been processed.
	*/

	for (cur = session->ssi.local ; cur != NULL ; cur = cur->next) {
		if (cur->type == 0) {
			char *group_name;
			struct buddy *buddy;
			struct bgroup *group;
			char *alias;

			if (cur->name == NULL)
				continue;

			group_name = aim_ssi_itemlist_findparentname(session->ssi.local,
							cur->name);

			alias = aim_ssi_getalias(session->ssi.local, group_name, cur->name);

			if (group_name == NULL)
				group_name = "orphans";

			buddy = buddy_find(priv, cur->name);
			if (buddy != NULL) {
				if (alias != NULL) {
					free(buddy->name);
					buddy->name = alias;
				}

				continue;
			}

			group = group_add(priv, group_name);
			buddy = buddy_add(priv, cur->name, group, 0);
			if (alias != NULL)
				buddy_alias(priv, buddy, alias, 0);
			else
				buddy_alias(priv, buddy, cur->name, 0);

			free(alias);
		} else if (cur->type == 2) {
			if (cur->name == NULL)
				continue;

			buddy_add_permit(priv, cur->name, 0);
		} else if (cur->type == 3) {
			/* Deny */
			if (cur->name == NULL)
				continue;

			buddy_add_block(priv, cur->name, 0);
		} else if (cur->type == 4) {
			u_int32_t privacy_mode = aim_ssi_getpermdeny(session->ssi.local);

			priv->buddy_pref->privacy_mode = privacy_mode;
		} else if (cur->type == 5) {
			u_int32_t report_idle = aim_ssi_getpresence(session->ssi.local);

			if (report_idle & 0x0000400)
				priv->buddy_pref->report_idle = 1;
			else
				priv->buddy_pref->report_idle = 0;
		}
	}

	if (!priv->connected) {
		aim_report_idle(priv, priv->buddy_pref->report_idle);

		aim_ssi_setpermdeny(session, fr->conn,
			priv->buddy_pref->privacy_mode, 0xffffffff);
	}

	aim_send_buddy_list(session, fr->conn, priv);
	aim_send_permit_list(session, fr->conn, priv);
	aim_send_block_list(session, fr->conn, priv);

	if (!priv->connected) {
		aim_clientready(session, fr->conn);

		priv->connected = 1;
		time(&priv->last_update);

		if (!do_event(priv->events, EVENT_SIGNON, priv->username))
			screen_status_msg("Logged in as %s successfully", priv->username);
	}

	return (1);
}

static FAIM_CB(aim_recv_popup) {
	/* Ignore this for now. */
	return (1);
}

static FAIM_CB(aim_recv_userinfo) {
	struct aim_priv *priv = session->aux_data;
	aim_userinfo_t *userinfo;
	u_int16_t info_type;
	char *encoding;
	char *result;
	va_list ap;
	char *profile = NULL;
	char *away_msg = NULL;
	u_int32_t idle_time = 0;
	u_int32_t online_since = 0;
	u_int32_t member_since = 0;
	u_int32_t warn_level;
	int print_all = 1;
	int text_len;

	va_start(ap, fr);
	userinfo = va_arg(ap, aim_userinfo_t *);
	info_type = (u_int16_t) va_arg(ap, unsigned int);
	encoding = va_arg(ap, char *);
	result = va_arg(ap, char *);
	text_len = va_arg(ap, int);
	va_end(ap);

	if (userinfo->present & AIM_USERINFO_PRESENT_ONLINESINCE)
		online_since = userinfo->onlinesince;

	if (userinfo->present & AIM_USERINFO_PRESENT_MEMBERSINCE)
		member_since = userinfo->membersince;

	if (userinfo->present & AIM_USERINFO_PRESENT_IDLE)
		idle_time = userinfo->idletime;

	warn_level = (float) userinfo->warnlevel / 10;

	switch (info_type) {
		case AIM_GETINFO_AWAYMESSAGE:
			if (do_event(priv->events, EVENT_RECV_AWAYMSG,
				userinfo->sn, priv->whois_hack,
				member_since, online_since, idle_time, warn_level,
				result, priv->username))
			{
				return (1);
			}

			if (result != NULL && *result != '\0') {
				char *temp = xmalloc(text_len + 1);

				xstrncpy(temp, result, text_len + 1);
				away_msg = strip_html(temp);
				free(temp);
			}

			break;

		case AIM_GETINFO_CAPABILITIES:
			break;

		default:
			if (do_event(priv->events, EVENT_RECV_PROFILE,
				userinfo->sn, member_since, online_since, idle_time,
				warn_level, result, priv->username))
			{
				return (1);
			}

			if (result != NULL && *result != '\0') {
				char *temp = xmalloc(text_len + 1);

				xstrncpy(temp, result, text_len + 1);
				profile = strip_html(temp);
				free(temp);
			}

			break;
	}

	if (info_type == AIM_GETINFO_AWAYMESSAGE && priv->whois_hack) {
		print_all = 0;
		priv->whois_hack = 0;
	}

	screen_print_info(userinfo->sn, profile, away_msg, warn_level,
		idle_time, online_since, member_since, print_all);

	if (away_msg != NULL)
		free(away_msg);

	if (profile != NULL)
		free(profile);

	return (1);
}

static FAIM_CB(aim_parse_login) {
	struct aim_priv *priv = (struct aim_priv *) session->aux_data;
	char *key;
	va_list ap;
	static struct client_info_s client_info = CLIENTINFO_AIM_KNOWNGOOD;

	va_start(ap, fr);
	key = va_arg(ap, char *);
	va_end(ap);

	aim_send_login(session, fr->conn,
		priv->username, priv->passwd, &client_info, key);

	return (1);
}

static FAIM_CB(aim_parse_authresp) {
	va_list ap;
	struct aim_authresp_info *authresp;
	aim_conn_t *bos_conn;
	struct aim_priv *priv = session->aux_data;

	pork_io_del(fr->conn);
	aim_conn_kill(session, &fr->conn);

	va_start(ap, fr);
	authresp = va_arg(ap, struct aim_authresp_info *);
	va_end(ap);

	if (authresp->errorcode || !authresp->bosip || !authresp->cookie) {
		switch (authresp->errorcode) {
			/* Incorrect nick / passwd */
			case 0x05:
				screen_err_msg("Error logging in: Incorrect username or password for the account %s",
					priv->username);

				if (priv->passwd != NULL) {
					memset(priv->passwd, 0, strlen(priv->passwd));
					free(priv->passwd);
					priv->passwd = NULL;
				}
				break;

			/* Suspended */
			case 0x11:
				screen_err_msg("Error logging in: The account '%s' has been suspended",
					priv->username);
				break;

			/* connecting too frequently */
			case 0x18:
				screen_err_msg("Error logging in: The account '%s' has been connecting too frequently",
					priv->username);
				break;

			/* client too old */
			case 0x1c:
				screen_err_msg("Error logging in: The AIM server says %s's client is too old",
					priv->username);
				break;
		}

		return (1);
	}

	bos_conn = aim_newconn(session, AIM_CONN_TYPE_BOS, authresp->bosip);
	if (bos_conn == NULL || bos_conn->fd == -1) {
		if (bos_conn != NULL)
			aim_conn_kill(session, &bos_conn);
		screen_err_msg("Unable to connect to the AIM BOS server");
		return (1);
	}

	pork_io_add(bos_conn->fd, IO_COND_RW, bos_conn, bos_conn, aim_conn_event);

	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_SPECIAL, AIM_CB_SPECIAL_CONNDEAD, aim_conn_dead, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_SPECIAL, AIM_CB_SPECIAL_CONNINITDONE, aim_recv_conn_complete, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_BOS, AIM_CB_BOS_RIGHTS, aim_recv_bos_rights, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_ACK, AIM_CB_ACK_ACK, NULL, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_GEN, AIM_CB_GEN_REDIRECT, aim_recv_redirect, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_LOC, AIM_CB_LOC_RIGHTSINFO, aim_recv_locrights, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_BUD, AIM_CB_BUD_RIGHTSINFO, aim_recv_buddy_rights, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_BUD, AIM_CB_BUD_ONCOMING, aim_recv_oncoming, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_BUD, AIM_CB_BUD_OFFGOING, aim_recv_offgoing, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_MSG, AIM_CB_MSG_INCOMING, aim_recv_msg, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_LOC, AIM_CB_LOC_ERROR, aim_recv_err_loc, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_MSG, AIM_CB_MSG_MISSEDCALL, aim_recv_missed, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_MSG, AIM_CB_MSG_CLIENTAUTORESP, aim_recv_clientauto, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_GEN, AIM_CB_GEN_RATECHANGE, aim_recv_rate_change, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_GEN, AIM_CB_GEN_EVIL, aim_recv_evil, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_LOK, AIM_CB_LOK_ERROR, aim_recv_search_error, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_LOK, 0x0003, aim_recv_search_reply, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_MSG, AIM_CB_MSG_ERROR, aim_recv_err_msg, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_MSG, AIM_CB_MSG_MTN, aim_recv_typing, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_LOC, AIM_CB_LOC_USERINFO, aim_recv_userinfo, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_MSG, AIM_CB_MSG_PARAMINFO, aim_recv_icbm_param_info, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_GEN, AIM_CB_GEN_ERROR, aim_recv_err_other, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_BUD, AIM_CB_BUD_ERROR, aim_recv_err_other, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_BOS, AIM_CB_BOS_ERROR, aim_recv_err_other, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_GEN, 0x001f, aim_recv_mem_request, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_GEN, AIM_CB_GEN_SELFINFO, aim_recv_selfinfo, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_POP, 0x0002, aim_recv_popup, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_SSI, AIM_CB_SSI_RIGHTSINFO, aim_ssi_recv_rights, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_SSI, AIM_CB_SSI_LIST, aim_ssi_recv_list, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_SSI, AIM_CB_SSI_NOLIST, aim_ssi_recv_list, 0);

	aim_sendcookie(session, bos_conn, authresp->cookielen, authresp->cookie);

	return (1);
}

/*
** No more connection handlers.
**
** Everything below is exported to other
** parts of the program.
*/

/*
** XXX: this is wrong, there should be a separate struct
** that holds information about the server, proxy, etc.
*/

int aim_login(struct aim_priv *priv) {
	aim_conn_t *auth_conn;
	char *server;

	screen_status_msg("Logging in as %s...", priv->username);

	if (priv->proxy != NULL) {
		aim_setupproxy(&priv->aim_session, priv->proxy,
			priv->proxy_user, priv->proxy_passwd);
	}

	server = priv->server;
	if (server == NULL)
		server = FAIM_LOGIN_SERVER;

	auth_conn = aim_newconn(&priv->aim_session, AIM_CONN_TYPE_AUTH, server);
	if (auth_conn == NULL) {
		screen_err_msg("Connection error while logging in as %s",
			priv->username);

		return (-1);
	}

	if (auth_conn->fd == -1) {
		if (auth_conn->status & AIM_CONN_STATUS_RESOLVERR)
			screen_err_msg("Unable to resolve authorizer hostname");
		else if (auth_conn->status & AIM_CONN_STATUS_CONNERR)
			screen_err_msg("Unable to connect to authorizer");
		else
			screen_err_msg("Error connecting to authorizer");

		aim_conn_kill(&priv->aim_session, &auth_conn);
		return (-1);
	}

	pork_io_add(auth_conn->fd, IO_COND_RW, auth_conn, auth_conn,
		aim_conn_event);

	aim_conn_addhandler(&priv->aim_session, auth_conn, AIM_CB_FAM_ATH, AIM_CB_ATH_AUTHRESPONSE, aim_parse_login, 0);
	aim_conn_addhandler(&priv->aim_session, auth_conn, AIM_CB_FAM_ATH, AIM_CB_ATH_LOGINRESPONSE, aim_parse_authresp, 0);

	aim_request_login(&priv->aim_session, auth_conn, priv->username);
	return (1);
}

int aim_setup(struct aim_priv *priv) {
	aim_session_t *session = &priv->aim_session;

	memset(session, 0, sizeof(*session));

	aim_session_init(session, 0, 0);
	aim_setdebuggingcb(session, aim_debug);
	aim_tx_setenqueue(session, AIM_TX_IMMEDIATE, NULL);

	session->aux_data = priv;

	return (0);
}

int aim_send_msg_auto(struct aim_priv *priv, char *dest, char *msg) {
	char *msg_html = text2html(msg);

	return (aim_send_im(&priv->aim_session, dest, 1, msg_html));
}

int aim_send_msg(struct aim_priv *priv, char *dest, char *msg) {
	int ret;
	char *msg_html = text2html(msg);

	ret = aim_send_im(&priv->aim_session, dest, 0, msg_html);

	if (priv->away_msg != NULL) {
		if (opt_get_bool(OPT_SEND_REMOVES_AWAY))
			aim_set_away(priv, NULL);
	}

	return (ret);
}

int aim_add_block(struct aim_priv *priv, char *target) {
	if (priv->aim_session.ssi.received_data == 0)
		return (-1);

	return (aim_ssi_adddeny(&priv->aim_session, priv->bos_conn, target));
}

int aim_remove_block(struct aim_priv *priv, char *target) {
	if (priv->aim_session.ssi.received_data == 0)
		return (-1);

	return (aim_ssi_adddeny(&priv->aim_session, priv->bos_conn, target));
}

int aim_add_permit(struct aim_priv *priv, char *target) {
	if (priv->aim_session.ssi.received_data == 0)
		return (-1);

	return (aim_ssi_addpermit(&priv->aim_session, priv->bos_conn, target));
}

int aim_remove_permit(struct aim_priv *priv, char *target) {
	if (priv->aim_session.ssi.received_data == 0)
		return (-1);

	return (aim_ssi_delpermit(&priv->aim_session, priv->bos_conn, target));
}

int aim_buddy_add(struct aim_priv *priv, struct buddy *buddy) {
	int ret;
	aim_session_t *session = &priv->aim_session;

	if (session->ssi.received_data == 0)
		return (-1);

	if (aim_ssi_itemlist_exists(session->ssi.local, buddy->nname))
		return (-1);

	ret = aim_ssi_addbuddy(session, priv->bos_conn, buddy->nname,
			buddy->group->name, buddy->name, NULL, NULL, 0);

	return (ret);
}

int aim_buddy_remove(struct aim_priv *priv, struct buddy *buddy) {
	int ret;

	if (priv->aim_session.ssi.received_data == 0)
		return (-1);

	ret = aim_ssi_delbuddy(&priv->aim_session, priv->bos_conn, buddy->nname,
			buddy->group->name);

	return (ret);
}

int aim_buddy_move(struct aim_priv *priv, struct buddy *buddy, char *old) {
	int ret;

	if (priv->aim_session.ssi.received_data == 0)
		return (-1);

	ret = aim_ssi_movebuddy(&priv->aim_session, priv->bos_conn, old,
			buddy->group->name, buddy->nname);

	return (ret);
}

int aim_buddy_alias(struct aim_priv *priv, struct buddy *buddy) {
	int ret;

	if (priv->aim_session.ssi.received_data == 0)
		return (-1);

	ret = aim_ssi_aliasbuddy(&priv->aim_session, priv->bos_conn,
			buddy->group->name, buddy->nname, buddy->name);

	return (ret);
}

int aim_warn(struct aim_priv *priv, char *target) {
	return (aim_send_warning(&priv->aim_session, priv->bos_conn, target, 0));
}

int aim_warn_anon(struct aim_priv *priv, char *target) {
	int ret;

	ret = aim_send_warning(&priv->aim_session, priv->bos_conn,
			target, AIM_WARN_ANON);

	return (ret);
}

int aim_set_away(struct aim_priv *priv, char *away_msg) {
	size_t len;
	int ret;
	char *msg_html;

	if (away_msg == NULL) {
		if (priv->away_msg != NULL)
			free(priv->away_msg);

		priv->away_msg = NULL;
		ret = aim_bos_setprofile(&priv->aim_session, priv->bos_conn,
				NULL, NULL, 0, NULL, "", 0, pork_caps);

		hash_destroy(&priv->autoreply);

		screen_status_msg("%s is longer away", priv->username);
		return (ret);
	}

	msg_html = text2html(away_msg);

	len = strlen(msg_html);
	if (len > priv->rights.max_away_len) {
		screen_err_msg("%s's away message is too long. The maximum length is %u characters",
			priv->username, priv->rights.max_away_len);
		return (-1);
	}

	if (priv->away_msg != NULL)
		free(priv->away_msg);
	priv->away_msg = xstrdup(away_msg);

	ret = aim_bos_setprofile(&priv->aim_session, priv->bos_conn,
			NULL, NULL, 0, "us-ascii", msg_html, len, pork_caps);

	hash_init(&priv->autoreply, 3, autoresp_compare_cb, autoresp_destroy_cb);

	screen_status_msg("%s is now marked as away", priv->username);
	return (ret);
}

int aim_keepalive(struct aim_priv *priv) {
	return (aim_flap_nop(&priv->aim_session, priv->bos_conn));
}

int aim_set_idle(struct aim_priv *priv, int idle_secs) {
	int ret = aim_bos_setidle(&priv->aim_session, priv->bos_conn, idle_secs);

	if (ret >= 0) {
		if (idle_secs > 0)
			priv->marked_idle = 1;
		else
			priv->marked_idle = 0;
	}

	return (ret);
}

int aim_set_profile(struct aim_priv *priv, char *profile) {
	int ret;
	size_t len;
	char *profile_html;

	if (profile == NULL) {
		ret = aim_bos_setprofile(&priv->aim_session,
				priv->bos_conn, NULL, "", 0, NULL, NULL, 0, pork_caps);

		return (ret);
	}

	profile_html = text2html(profile);

	len = strlen(profile_html);
	if (len > priv->rights.max_profile_len) {
		screen_err_msg("%s's profile is too long. The maximum length is %u characters",
			priv->username, priv->rights.max_profile_len);
		return (-1);
	}

	ret = aim_bos_setprofile(&priv->aim_session, priv->bos_conn,
		"us-ascii", profile_html, len, NULL, NULL, 0, pork_caps);

	return (ret);
}

int aim_get_away_msg(struct aim_priv *priv, char *buddy) {
	int ret;

	ret = aim_getinfo(&priv->aim_session,
			priv->bos_conn, buddy, AIM_GETINFO_AWAYMESSAGE);

	priv->whois_hack = 0;
	return (ret);
}

int aim_whois(struct aim_priv *priv, char *buddy) {
	int ret;

	/*
	** There's no facility provided to us that allows us
	** to fetch a user's profile and away message at the same
	** time.  Instead, we have to request both separately.
	** The 'whois_hack' variable is set so that when the
	** routine that normally would print the away message
	** (the same one that prints profile info) is called
	** it knows only to print the away message and nothing else
	** for this particular request.  Ordinarily, information
	** like signon time, idle time, username, etc is printed
	** when we receive a response to a "get away message"
	** request.
	*/

	ret = aim_getinfo(&priv->aim_session,
			priv->bos_conn, buddy, AIM_GETINFO_GENERALINFO);

	aim_getinfo(&priv->aim_session,
		priv->bos_conn, buddy, AIM_GETINFO_AWAYMESSAGE);

	priv->whois_hack = 1;
	return (ret);
}

void aim_kill_all_conn(struct aim_priv *priv) {
	aim_conn_t *conn = priv->aim_session.connlist;
	dlist_t *cur;

	if (priv->connected) {
		aim_rxdispatch(&priv->aim_session);
		aim_tx_flushqueue(&priv->aim_session);
	}

	while (conn != NULL) {
		pork_io_del(conn);
		conn = conn->next;
	}

	cur = priv->chat_list;
	while (cur != NULL) {
		dlist_t *next = cur->next;

		aim_free_chat(priv, cur);
		cur = next;
	}

	aim_session_kill(&priv->aim_session);
}

int aim_privacy_mode(struct aim_priv *priv, int mode) {
	if (mode >= 0 && mode <= 5) {
		priv->buddy_pref->privacy_mode = mode;
		aim_ssi_setpermdeny(&priv->aim_session, priv->bos_conn,
			priv->buddy_pref->privacy_mode, 0xffffffff);
	}

	return (priv->buddy_pref->privacy_mode);
}

int aim_report_idle(struct aim_priv *priv, int mode) {
	u_int32_t report_idle = aim_ssi_getpresence(priv->aim_session.ssi.local);
	int ret;

	if (mode != 0)
		report_idle |= 0x0000400;
	else
		report_idle &= ~0x0000400;

	priv->buddy_pref->report_idle = ((report_idle & 0x0000400) != 0);
	ret = aim_ssi_setpresence(&priv->aim_session, priv->bos_conn, report_idle);
	return (ret);
}

int aim_join_chatroom(struct aim_priv *priv, char *name, int exchange) {
	aim_conn_t *chatnav_conn;
	int ret;

	if (exchange < 4)
		exchange = 4;
	else if (exchange > 16)
		exchange = 16;

	chatnav_conn = priv->chatnav_conn;
	if (chatnav_conn == NULL) {
		struct chatroom_info *info = xcalloc(1, sizeof(*info));

		info->name = xstrdup(name);
		info->exchange = (u_int16_t) exchange;

		priv->chat_create_list = dlist_add_head(priv->chat_create_list, info);
		aim_reqservice(&priv->aim_session, priv->bos_conn,
			AIM_CONN_TYPE_CHATNAV);

		return (0);
	}

	ret = aim_chatnav_createroom(&priv->aim_session, chatnav_conn,
			name, exchange);
	return (ret);
}

dlist_t *aim_find_chat(struct aim_priv *priv, const char *name, int exchange) {
	dlist_t *cur;

	cur = priv->chat_list;
	while (cur != NULL) {
		struct chatroom *chat = cur->data;

		if (!strcasecmp(name, chat->title) && exchange == chat->exchange)
			return (cur);

		cur = cur->next;
	}

	return (NULL);
}

dlist_t *aim_find_chat_name(struct aim_priv *priv, char *name) {
	char *p;
	char *orig;
	int exchange = AIM_DEFAULT_CHAT_EXCHANGE;
	dlist_t *ret;

	p = strchr(name, '/');
	orig = p;
	if (p != NULL) {
		*p++ = '\0';

		if (str_to_int(p, &exchange) != 0) {
			*orig = '/';
			return (NULL);
		}
	}

	ret = aim_find_chat(priv, name, exchange);

	if (orig != NULL)
		*orig = '/';

	return (ret);
}

int aim_leave_chatroom(struct aim_priv *priv, char *name, int exchange) {
	dlist_t *cur;

	cur = aim_find_chat(priv, name, exchange);
	if (cur == NULL)
		return (-1);

	aim_free_chat(priv, cur);
	return (0);
}

int aim_chat_send(struct aim_priv *priv, char *msg, struct chatroom *chat) {
	char *msg_html;
	int msg_len;
	int visible_len;
	char *visible;
	int ret;

	if (msg == NULL)
		return (-1);

	msg_html = text2html(msg);

	msg_len = strlen(msg_html);

	if (msg_len > chat->max_msg_len)
		return (-1);

	visible = strip_html(msg);
	if (visible == NULL)
		return (-1);

	visible_len = strlen(visible);
	free(visible);

	if (visible_len > chat->max_visible_len)
		return (-1);

	ret = aim_chat_send_im(&priv->aim_session, chat->conn,
			AIM_CHATFLAGS_NOREFLECT, msg_html, msg_len);

	return (ret);
}

int aim_chat_send_name(	struct aim_priv *priv,
						char *msg,
						char *name,
						int exchange)
{
	dlist_t *cur = aim_find_chat(priv, name, exchange);

	if (cur == NULL)
		return (-1);

	return (aim_chat_send(priv, msg, cur->data));
}

int aim_chat_send_invite(	struct aim_priv *priv,
							char *dest,
							char *msg,
							struct chatroom *chat)
{
	int ret;

	if (msg == NULL)
		msg = "";

	ret = aim_chat_invite(&priv->aim_session, priv->bos_conn, dest, msg,
			chat->exchange, chat->name, chat->exchange);

	return (ret);
}

int aim_chat_ignore(char *user, struct chatroom *chat) {
	dlist_t *node;
	struct buddy *buddy;

	node = dlist_find(chat->user_list, user, chat_user_find_cb);
	if (node == NULL)
		return (-1);

	buddy = node->data;
	buddy->ignore = 1;

	return (0);
}

int aim_chat_unignore(char *user, struct chatroom *chat) {
	dlist_t *node;
	struct buddy *buddy;

	node = dlist_find(chat->user_list, user, chat_user_find_cb);
	if (node == NULL)
		return (-1);

	buddy = node->data;
	buddy->ignore = 0;

	return (0);
}

void aim_chat_print_users(struct imwindow *win, struct chatroom *chat) {
	dlist_t *cur = chat->user_list;
	char buf[2048];
	size_t len = sizeof(buf);
	size_t i = 0;
	int ret;
	struct buddy *buddy;

	if (cur == NULL) {
		screen_win_msg(win, 0, 0, 1,
			"--%%m--%%M--%%x No %%Cu%%csers%%x in %%C%s/%u%%x (%%W%s%%x)",
			chat->title, chat->exchange, chat->name);

		return;
	}

	screen_win_msg(win, 0, 0, 1,
		"--%%m--%%M--%%x %u %%Cu%%csers%%x in %%C%s/%u%%x (%%W%s%%x)",
		chat->num_users, chat->title, chat->exchange, chat->name);

	while (cur != NULL) {
		buddy = cur->data;

		if (buddy->ignore)
			ret = snprintf(&buf[i], len, "[%%R%s%%x] ", buddy->name);
		else
			ret = snprintf(&buf[i], len, "[%%B%s%%x] ", buddy->name);

		len -= ret;
		i += ret;

		cur = cur->next;
	}

	if (i > 0 && buf[i - 1] == ' ')
		buf[i - 1] = '\0';

	screen_win_msg(win, 0, 0, 1, "%s", buf);
}

void aim_update_buddy(struct buddy *buddy, aim_userinfo_t *userinfo) {
	buddy->signon_time = userinfo->onlinesince;
	buddy->warn_level = (float) userinfo->warnlevel / 10;

	if (userinfo->present & AIM_USERINFO_PRESENT_FLAGS) {
		if (userinfo->flags & AIM_FLAG_UNCONFIRMED)
			buddy->type |= USER_TYPE_UNCONFIRMED;
		if (userinfo->flags & AIM_FLAG_FREE)
			buddy->type |= USER_TYPE_AIM;
		if (userinfo->flags & AIM_FLAG_AOL)
			buddy->type |= USER_TYPE_AOL;
		if (userinfo->flags & AIM_FLAG_WIRELESS)
			buddy->type |= USER_TYPE_WIRELESS;
		if (userinfo->flags & AIM_FLAG_ADMINISTRATOR)
			buddy->type |= USER_TYPE_ADMIN;
		if (userinfo->flags & AIM_FLAG_ACTIVEBUDDY)
			buddy->type |= USER_TYPE_ACTIVEBUDDY;
	}

	if (userinfo->present & AIM_USERINFO_PRESENT_SESSIONLEN)
		buddy->idle_time = userinfo->idletime;

	buddy->status = STATUS_ACTIVE;

	if (buddy->idle_time > 0)
		buddy->status = STATUS_IDLE;

	if (userinfo->flags & AIM_FLAG_AWAY)
		buddy->status = STATUS_AWAY;
}
